/*
 * Copyright 2014, 2016 Peter Olsson
 * 
 * This file is part of Brum Brum Rally.
 * 
 * Brum Brum Rally is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Brum Brum Rally is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "Hqx.h"
#include "hqx/hqx.h"
#include <cstring>

Hqx::Hqx()
{
	hqxInit();
}

bool Hqx::isValidZoom(int zoom) const
{
	while ((zoom % 4) == 0)
	{
		zoom /= 4;
	}
	if ((zoom % 2) == 0)
	{
		zoom /= 2;
	}
	if ((zoom % 3) == 0)
	{
		zoom /= 3;
	}
	return zoom == 1;
}

bool Hqx::isValidBpp(int bpp) const
{
	return bpp == 32;
}

void Hqx::renderStretch(const SDL_Rect& srcArea_, SDL_Rect& dstArea, SDL_Surface* source, SDL_Surface* destination)
{
	SDL_Rect srcArea = srcArea_;
	expand(srcArea, dstArea, 1);
	
	int x = 0;
	int y = 0;
	
	if (!buffer)
	{
		x = -zoom * srcArea.x + dstArea.x + dstArea.w - zoom * srcArea.w;
		y = -zoom * srcArea.y + dstArea.y + dstArea.h - zoom * srcArea.h;
	}
	
	renderZoom(srcArea, zoom, x, y, source, (buffer ? buffer : destination));
	
	srcArea.x = dstArea.x + dstArea.w - zoom * srcArea.w;
	srcArea.y = dstArea.y + dstArea.h - zoom * srcArea.h;
	srcArea.w *= zoom;
	srcArea.h *= zoom;

	stretchArea(srcArea, dstArea, x, y, (buffer ? buffer : destination), destination);
}

void Hqx::renderZoom(const SDL_Rect& srcArea_, SDL_Rect& dstArea, SDL_Surface* source, SDL_Surface* destination)
{
	SDL_Rect srcArea = srcArea_;
	if (srcArea.x > 0)
	{
		--srcArea.x;
		++srcArea.w;
	}
	if (srcArea.y > 0)
	{
		--srcArea.y;
		++srcArea.h;
	}
	if (srcArea.x + srcArea.w < GAME_WIDTH)
	{
		++srcArea.w;
	}
	if (srcArea.y + srcArea.h < GAME_HEIGHT)
	{
		++srcArea.h;
	}
	dstArea.x = screenArea.x + srcArea.x * zoom;
	dstArea.y = screenArea.y + srcArea.y * zoom;
	dstArea.w = srcArea.w * zoom;
	dstArea.h = srcArea.h * zoom;
	
	renderZoom(srcArea, zoom, screenArea.x, screenArea.y, source, destination);
}

void Hqx::renderZoom(const SDL_Rect& srcArea, int zoom, int x, int y,  SDL_Surface* source, SDL_Surface* destination)
{
	const int savedZoom = zoom;
	
	int sx = srcArea.x;
	int sy = srcArea.y;
	int w = srcArea.w;
	int h = srcArea.h;
	int dx;
	int dy;
	
	while (zoom > 4 && (zoom % 4) == 0)
	{
		dx = x + (srcArea.x + srcArea.w) * savedZoom - 4 * w;
		dy = y + (srcArea.y + srcArea.h) * savedZoom - 4 * h;
		
		hq4x_32_rb(source->pixels, source->pitch, destination->pixels, destination->pitch, dx, dy, sx, sy, w, h);
		
		w *= 4;
		h *= 4;
		sx = dx;
		sy = dy;
		source = destination;
		zoom /= 4;
	}
	
	if (zoom > 4 && (zoom % 2) == 0)
	{
		dx = x + (srcArea.x + srcArea.w) * savedZoom - 2 * w;
		dy = y + (srcArea.y + srcArea.h) * savedZoom - 2 * h;
		
		hq2x_32_rb(source->pixels, source->pitch, destination->pixels, destination->pitch, dx, dy, sx, sy, w, h);
		
		w *= 2;
		h *= 2;
		sx = dx;
		sy = dy;
		source = destination;
		zoom /= 2;
	}
	dx = x + srcArea.x * savedZoom;
	dy = y + srcArea.y * savedZoom;
	
	switch (zoom)
	{
	case 1:
	{
		SDL_Rect rect;
		rect.x = srcArea.x + x;
		rect.y = srcArea.y + y;
		SDL_BlitSurface(source, const_cast<SDL_Rect*>(&srcArea) , destination, &rect);
		break;
	}
	case 2:
		hq2x_32_rb(source->pixels, source->pitch, destination->pixels, destination->pitch, dx, dy, sx, sy, w, h);
		break;
	case 3:
		hq3x_32_rb(source->pixels, source->pitch, destination->pixels, destination->pitch, dx, dy, sx, sy, w, h);
		break;
	case 4:
		hq4x_32_rb(source->pixels, source->pitch, destination->pixels, destination->pitch, dx, dy, sx, sy, w, h);
		break;
	}
}
